#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/socket.h>
#ifndef USE_NETLIB
#include <net/ethernet.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#ifdef __linux__
#include <netinet/ether.h>
#endif
#include <arpa/inet.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#else
#include <netlib.h>
#endif

#include "pktlib.h"
#include "pktbuf.h"
#include "pcap.h"

void usage()
{
  printf("[-i <interface>]\n");
  exit(0);
}

struct pktbuf_option {
  pktif_t rcvif;
};

static int finished = 0;

int proc_arp(pktbuf_t pktbuf, struct header_param *param)
{
  struct in_addr in;

  printf("ARP:\n");
  printf("\theader_size:\t\t%d\n", param->arp.header_size);
  printf("\tpayload_size:\t\t%d\n", param->arp.payload_size);
  printf("\thardware:\t\t0x%04x\n", param->arp.hardware);
  printf("\tprotocol:\t\t0x%04x\n", param->arp.protocol);
  printf("\thardware_addr_size:\t%d\n", param->arp.hardware_addr_size);
  printf("\tprotocol_addr_size:\t%d\n", param->arp.protocol_addr_size);
  printf("\toperation:\t\t0x%04x\n", param->arp.operation);

  if (param->arp.hardware == ARPHRD_ETHER) {
    printf("\tsender_hardware_addr:\t%s\n",
	   ether_ntoa((struct ether_addr *)param->arp.sender_hardware_addr));
    printf("\ttarget_hardware_addr:\t%s\n",
	   ether_ntoa((struct ether_addr *)param->arp.target_hardware_addr));
  }

  if (param->arp.protocol == ETHERTYPE_IP) {
    in.s_addr = htonl(param->arp.sender_protocol_addr);
    printf("\tsender_protocol_addr:\t%s\n", inet_ntoa(in));
    in.s_addr = htonl(param->arp.target_protocol_addr);
    printf("\ttarget_protocol_addr:\t%s\n", inet_ntoa(in));
  }

  return 0;
}

int proc_icmp(pktbuf_t pktbuf, struct header_param *param)
{
  printf("ICMP:\n");
  printf("\theader_size:\t%d\n", param->icmp.header_size);
  printf("\tpayload_size:\t%d\n", param->icmp.payload_size);
  printf("\ttype:\t\t%d\n", param->icmp.type);
  printf("\tcode:\t\t%d\n", param->icmp.code);
  return 0;
}

int proc_tcp(pktbuf_t pktbuf, struct header_param *param)
{
  printf("TCP:\n");
  printf("\theader_size:\t%d\n", param->tcp.header_size);
  printf("\tpayload_size:\t%d\n", param->tcp.payload_size);
  printf("\tsrc_port:\t%d\n", param->tcp.src_port);
  printf("\tdst_port:\t%d\n", param->tcp.dst_port);
  printf("\tseq_number:\t0x%08x\n", param->tcp.seq_number);
  printf("\tack_number:\t0x%08x\n", param->tcp.ack_number);
  printf("\toffset:\t\t0x%02x\n", param->tcp.offset);
  printf("\tflags:\t\t0x%02x\n", param->tcp.flags);
  printf("\twindow:\t\t0x%04x\n", param->tcp.window);
  return 0;
}

int proc_udp(pktbuf_t pktbuf, struct header_param *param)
{
  printf("UDP:\n");
  printf("\theader_size:\t%d\n", param->udp.header_size);
  printf("\tpayload_size:\t%d\n", param->udp.payload_size);
  printf("\tsrc_port:\t%d\n", param->udp.src_port);
  printf("\tdst_port:\t%d\n", param->udp.dst_port);
  return 0;
}

int proc_ip(pktbuf_t pktbuf, struct header_param *param)
{
  struct in_addr in;

  printf("IP:\n");
  printf("\theader_size:\t%d\n", param->ip.header_size);
  printf("\tpayload_size:\t%d\n", param->ip.payload_size);
  printf("\tid:\t\t0x%04x\n", param->ip.id);
  printf("\tttl:\t\t%d\n", param->ip.ttl);
  printf("\tprotocol:\t%d\n", param->ip.protocol);

  in.s_addr = htonl(param->ip.src_ipaddr);
  printf("\tsrc_ipaddr:\t%s\n", inet_ntoa(in));
  in.s_addr = htonl(param->ip.dst_ipaddr);
  printf("\tdst_ipaddr:\t%s\n", inet_ntoa(in));

  pktbuf_delete_header(pktbuf, param->ip.header_size);

  switch (param->ip.protocol) {
  case IPPROTO_ICMP: proc_icmp(pktbuf, param); break;
  case IPPROTO_UDP:  proc_udp(pktbuf, param);  break;
  case IPPROTO_TCP:  proc_tcp(pktbuf, param);  break;
  default:
    break;
  }

  pktbuf_add_header(pktbuf, param->ip.header_size);

  return 0;
}

int proc_ethernet(pktbuf_t pktbuf, struct header_param *param)
{
  printf("Ethernet:\n");
  printf("\theader_size:\t%d\n", param->ethernet.header_size);
  printf("\tpayload_size:\t%d\n", param->ethernet.payload_size);
  printf("\tdst_macaddr:\t%s\n", ether_ntoa((struct ether_addr *)param->ethernet.dst_macaddr));
  printf("\tsrc_macaddr:\t%s\n", ether_ntoa((struct ether_addr *)param->ethernet.src_macaddr));
  printf("\ttype:\t\t0x%04x\n", param->ethernet.type);
  printf("\tvid:\t\t%d\n", param->ethernet.vid);

  pktbuf_delete_header(pktbuf, ETHER_HDR_LEN);

  switch (param->ethernet.type) {
  case ETHERTYPE_ARP: proc_arp(pktbuf, param); break;
  case ETHERTYPE_IP:  proc_ip(pktbuf, param);  break;
  default:
    break;
  }

  pktbuf_add_header(pktbuf, ETHER_HDR_LEN);

  return 0;
}

int main(int argc, char *argv[])
{
  pktif_t pktif, p;
  int r, ch;
  pktbuf_t pktbuf;
  int type, size, count = 0;
  struct header_param param;
  FILE *fp = NULL;
  struct timeval *t;
  pktbuf_t packets = NULL;

  pktif = NULL;

  while ((ch = getopt(argc, argv, "i:f:")) != -1) {
    switch (ch) {
    case 'i':
      /* 󥿡ե򥪡ץ󤹤(ȯѥåȤϼʤ) */
      pktif = pktif_open(optarg, PKTIF_OPEN_FLAG_RECV_NOTSENT, 0);
      break;
    case 'f':
      fp = fopen(optarg, "w");
      break;
    case '?':
    default:
      usage();
    }
  }

  pktbuf_init(sizeof(struct pktbuf_option));

  while (!finished) {
    pktif = pktlib_iflist_select(packets ? 0 : -1); /* ʣ󥿡եԤ */
    if (pktif) {
      pktbuf = pktbuf_recv(pktif, 0); /* ѥåȤ */
      if (pktbuf) {
	((struct pktbuf_option *)pktbuf_get_option(pktbuf))->rcvif = pktif;
      }
      pktbuf_enqueue(&packets, pktbuf);
      continue;
    }

    pktbuf = pktbuf_dequeue(&packets);
    if (!pktbuf)
      continue;

    pktif = ((struct pktbuf_option *)pktbuf_get_option(pktbuf))->rcvif;

    size = pktbuf_get_size(pktbuf);
    t = pktbuf_get_time(pktbuf);

    type = pktif_get_linktype(pktif);
    if (type == DLT_EN10MB) {
      r = pktbuf_analyze_ethernet(pktbuf, &param);
      if (r >= 0) {
	printf("-- %d --\n", ++count);
	printf("SIZE: %d\n", size);
	proc_ethernet(pktbuf, &param);
	printf("==\n");
	fflush(stdout);
      }
    }

    if (fp) {
      pkt_pcap_write(fp, pktbuf_get_header(pktbuf), size, type, size, t);
    }

    /* ѥåȤ */
    for (p = pktlib_iflist_get_list(); p; p = pktif_get_next(p)) {
      if (p == pktif)
	continue;
      pktbuf_send(p, pktbuf_clone(pktbuf));
    }

    pktbuf_destroy(pktbuf);
  }

  if (fp)
    fclose(fp);

  return 0;
}
